/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg004HeuschBoesefeldt
 *
 * de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg004HeuschBoesefeldt is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg004HeuschBoesefeldt is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg004HeuschBoesefeldt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg004HeuschBoesefeldt.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg004HeuschBoesefeldt ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg004HeuschBoesefeldt wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7.conversion.nw.vrz;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.kex.tls.osi7.*;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.hexdump.HexDumper;

import java.util.ArrayList;
import java.util.List;

/**
 * Umsetungsmodul für die FG 4, wird für die WanCom-Kopplung der ZDF basierten VRZ Leverkusen und VRZ Recklinghausen mit
 * der DaV basierten VRZ NRW verwendet.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */

public class Fg004HeuschBoesefeldt extends FgAllHeuschBoesefeldt {
// ------------------------------------------------------------------------
// VARIABLENDEKLARATIONEN
// ------------------------------------------------------------------------

// Sonstige ---------------------------------------------------------------
	/** DebugLogger für Debug-Ausgaben */
	private static final Debug debug = Debug.getLogger();

	private De _de;
// Private ----------------------------------------------------------------
	/**
	 * Tabelle, in der das Verhalten beim Empfang von Attributgruppen eingetragen wird. Jede Tabellenzeile hat folgenden
	 * Aufbau:
	 * <pre>
	 * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
	 * z.B.
	 * "atg.tlsGloKanalSteuerung", "asp.parameterSoll", "1", "29", "2", "typ.deLve"
	 * </pre>
	 * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine
	 * Empfangsanmeldung als Empfänger.</li> <li> In der Methode {@link #update} wird für jede zur
	 * Attributgruppen/Aspket-Kombination passende Kombination eine Methode aufgerufen, deren Namen aus dem Tabelleneintrag
	 * wie folgt gebildet wird (hier für den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspParameterSoll(ResultData)</code><br>
	 * Dabei enthält <code>ResultData</code> die Daten der zu verarbeitenden ATG/ASP-Kombination.</li> <li> Für jede
	 * Attributguppen/Aspekt Kombination erfolgt eine Empfangsanmeldung als Empfänger. Sind mehrere passende Eintrage
	 * vorhanden, werden entsprechend mehrere Methodenaufrufe durchgeführt. Dadurch können mehrere ATG/ASP-Kombinationen in
	 * einen De-Block abgebildet werden (z.B. protokollnahe TLS-Attributgruppe und Attributgruppe eines Fachmodells):</li>
	 * <li> Ein Eintrag von "255" entspricht dabei einer "Wildcard" (passt immer). </ul>
	 */
	private final Object[][] davToTls = new Object[][]{
			// Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp
			{"atg.tlsGloDeBlockAbruf", "asp.tlsAbruf", "255", "255", "255", "typ.deWzg"},
			{"atg.tlsWzgWvzStellZustand", "asp.tlsVorgabe", "4", "48", "81", "typ.deWzg"},
			{"atg.tlsWzgWechselText", "asp.tlsVorgabe", "4", "50", "81", "typ.deWzg"},
	};

	/**
	 * Tabelle, in der das Verhalten beim Empfang von TLS De-Blöcken eingetragen wird. Jede Tabellenzeile hat folgenden
	 * Aufbau:
	 * <pre>
	 * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
	 * z.B.
	 * "atg.tlsGloKanalSteuerung", "asp.tlsAntwort", "1", "29", "2", "typ.deLve"
	 * </pre>
	 * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine
	 * Sendeanmeldung als Quelle.</li> <li> In der Methode {@link #processDeBlock} wird für jede zum De-Block passende
	 * Fg/Id/Typ-Kombination eine Methode aufgerufen, deren Namen aus dem Tabelleneintrag wie folgt gebildet wird (hier für
	 * den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspTlsAntwort(DeBlock, Data,
	 * DataDescription)</code><br> Dabei enthält der <code>deBlock</code> die Daten des zu verarbeitenden De-Blocks und
	 * <code>data</code> ein Datenobjekt der passenden Attributgruppen/Aspekt-Kombination, welchem direkt die Inhalte des
	 * De-Blocks zugewiesen werden können.<code>DataDescription</code> enthält die passende {@link
	 * DataDescription} für die zu sendenden Daten.</li> <li> Für jede Attributguppen/Aspekt
	 * Kombination erfolgt eine Sendeanmeldung als Quelle. Sind mehrere passende Eintrage vorhanden, werden entsprechend
	 * mehrere Methodenaufrufe durchgeführt. Dadurch kann ein De-Block mehrfach ausgewert und an verschiedene
	 * Attributgruppen weitergerleitet werden (z.B. protokollnahe TLS-Attributgruppe und Attributgruppe eine
	 * Fachmodells):</li> <li> Ein Eintrag von "255" entspricht dabei einer "Wildcard" (passt immer). </ul>
	 */
	private final Object[][] tlsToDav = new Object[][]{
			// Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp
			{"atg.tlsGloDeFehler", "asp.tlsAntwort", "4", "1", "1", "typ.deWzg"},
			{"atg.tlsGloKanalSteuerung", "asp.parameterIst", "255", "29", "2", "typ.deWzg"},
			{"atg.tlsGloGeographischeKenndaten", "asp.tlsZustand", "255", "36", "3", "typ.deWzg"},
			{"atg.tlsWzgNichtDarstellbareWVZ", "asp.tlsAntwort", "4", "2", "1", "typ.deWzg"},
			{"atg.tlsWzgDefekteLampen", "asp.tlsAntwort", "4", "3", "1", "typ.deWzg"},
			{"atg.tlsWzgBetriebsArt", "asp.parameterIst", "4", "17", "2", "typ.deWzg"},
			{"atg.tlsWzgWvzGrundEinstellung", "asp.parameterIst", "4", "32", "3", "typ.deWzg"},
			{"atg.tlsWzgWvzGrundEinstellung", "asp.parameterIst", "4", "32", "35", "typ.deWzg"},
			{"atg.tlsWzgGrundProgramm", "asp.parameterIst", "4", "42", "3", "typ.deWzg"},
			{"atg.tlsWzgWvzStellZustand", "asp.tlsZustand", "4", "48", "5", "typ.deWzg"},
			{"atg.tlsWzgHelligkeit", "asp.tlsZustand", "4", "49", "5", "typ.deWzg"},
			{"atg.tlsWzgWechselText", "asp.tlsZustand", "4", "50", "5", "typ.deWzg"},
			{"atg.tlsWzgStellZustand", "asp.tlsZustand", "4", "55", "5", "typ.deWzg"},
			{"atg.tlsWzgBetriebsArt", "asp.tlsVorgabeBedienung", "4", "17", "80", "typ.deWzg"}
	};

	/**
	 * Gültiger WvzCode, der beim Einschalten oder als Rückmeldung verwendet wurde. Da sich eine Reihe von
	 * Softwareentwicklern für die SM/EAK-Software bei den Firmen auf eine etwas seltsame Interpretation der TLS im Falle
	 * des Ausschaltens von WVZ-Anzeigen versteift (1), wird bis zur verbindlichen Festlegung einer vernünftigen
	 * Interpretation ein WVZ beim Ausschalten jeweils mit dem letzten verschickten oder empfangenen WVZ-Code einer
	 * Einschaltsequenz ausgeschaltet. (1) Das Schalten einer WVZ-Anzeige erfolgt über zwei Steuerbyte: dem eigentlichen
	 * Befehl ob ein- oder auszuschalten ist und dem darzustellenden WVZ-Code. Letzterer macht logischerweise nur Sinn beim
	 * Einschalten. Einige Entwickler versteifen sich aber darauf, dass auch beim Ausschalten ein gültiger Code verwendet
	 * werden muss und schalten das WZG ansonsten nicht aus (obwohl kein Mensch erklären kann, warum hier so ein unnötiger
	 * Programmieraufwand getrieben wird, der zudem einen sicheren Betrieb erheblich erschwert...Im Zweifelsfall kann
	 * nämlich eine Fehlanzeige nicht ohne weiteres über eine Mánuelle Schaltung ausgeschaltet werden, weil die Liste der
	 * gültigen Anzeigecode nicht bekannt ist!!).
	 */
	private int _validWvzCode = 0;// initial auf 0 setzen

// ------------------------------------------------------------------------
// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
// ------------------------------------------------------------------------

	/**
	 * Erzeugt ein Objekt vom Typ Fg004Default
	 *
	 * @param daf DaV
	 * @param eak Eak
	 * @param de De
	 */
	public Fg004HeuschBoesefeldt(ClientDavInterface daf, Eak eak, De de) {
		super(daf, eak, de);
		subscribeAllData(tlsToDav, davToTls);
		_de = de;
	}


// ------------------------------------------------------------------------
// SONSTIGE METHODEN
// ------------------------------------------------------------------------

	/**
	 * FG 4 : Typ: 1 Id: 1 Name: DE-Fehler (Antwort). Ergänzt globale Behandlung De-Fehler in {@link
	 * #Fg255Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort}. Beim Empfang einer De-Gutmeldung wird die De initialisiert, indem
	 * <ul> <li>WZG-kanalsteuerung {@link #Fg255Typ029Id002AtgTlsGloKanalSteuerungAspParameterSoll} TBD: Folgende
	 * Einstellung bei FG 4 nach DE-Gutmeldung versenden (aber wie werden die verschiedenen Bauformaen, TLS-Stände
	 * konfiguriert ??? {"atg.tlsWzgBetriebsArt", "asp.parameterSoll", "4", "17", "2", "typ.deWzg"},
	 * {"atg.tlsWzgWvzGrundEinstellung", "asp.parameterSoll", "4", "32", "3", "typ.deWzg"}, {"atg.tlsWzgGrundEinstellung",
	 * "asp.parameterSoll", "4", "33", "3", "typ.deWzg"}, {"atg.tlsWzgGrundProgramm", "asp.parameterSoll", "4", "42", "3",
	 * "typ.deWzg"}, {"atg.tlsWzgCodeListe", "asp.parameterSoll", "4", "43", "3", "typ.deWzg"}, </li></ul> versendet werden.
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = super.Fg255Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort(deBlock, data, dataDescription);

		IntegerValueState DEFehlerStatus = data.getUnscaledValue("DEFehlerStatus").getState();
		if (DEFehlerStatus != null) {
			debug.fine("DEFehlerStatus: " + DEFehlerStatus.getName());
			if (DEFehlerStatus.getName().equals("ok")) {
				// Bei der Kommunikation mit der VRZ werden die Daten zu folgenden ATG/ASP-Kombinationen abgefragt,
				// damit insbesondere die dWiSta-Tafeln in dMotion korrekte Inhalte anzeigen:
				// Problem: Bei der Kommnikation über WAN-COM mit der HB-VRZ werden nach einer DE-Gutmeldung
				// keine Stellzustandsdaten etc. von der HB-VRZ an die neue VRZ gesendet.
				//

				// ATG                               ASP                 TYP   ID   Kommentar
				// für Symbole:
				//
				// atg.tlsGloDeFehler                asp.tlsAntwort        1    1   ist vorhanden (keine Abfrage)
				// atg.tlsWzgNichtDarstellbareWVZ    asp.tlsAntwort        2    1   Nicht darstellbare WVZ
				// atg.tlsWzgWvzStellZustand         asp.tlsZustand       48    5   WVZ Stellzustand
				// atg.tlsGloKanalSteuerung          asp.parameterIst     29    2   Kanal-Steuerung
				//
				// für Wechseltexte:
				//
				// atg.tlsGloDeFehler                asp.tlsAntwort        1    1   ist vorhanden (keine Abfrage)
				// atg.tlsWzgGestörteTextPositionen  asp.tlsAntwort        4    1   Gestörte Textpositionen
				// atg.tlsWzgWechselText             asp.tlsZustand       50    5   Wechseltext
				// atg.tlsGloKanalSteuerung          asp.parameterIst     29    2   Kanal-Steuerung

				// Typ/ID Kombinationen der FG 4, die bei DE-Gutmeldungen abgefragt werden müssen
				int[][] missingData = new int[][]{// {1, 1}, // De-Fehler
				                                  {2, 1}, // Nicht darstellbare WVZ
				                                  // {3, 1}, // Defekte Lampen
				                                  {4, 1}, // Gestörte Textpositionen
				                                  //{14, 1}, // Ergänzende DE-Fehler
				                                  //{17, 2}, // Betriebsart
				                                  {29, 2}, // Kanal-Steuerung
				                                  //{32, 3}, // WZG Grundeinstellung
				                                  //{33, 3}, // Grundeinstellung
				                                  //{36, 3}, // Geografische Kenndaten
				                                  //{42, 3}, // Grundprogramm
				                                  //{43, 3}, // Codeliste
				                                  //{44, 3}, // Code Definition
				                                  {48, 5}, // WVZ Stellzustand
				                                  //{49, 5}, // Helligkeit
				                                  {50, 5}, // Wechseltext
				                                  //{55, 5}, // Stellzustand
				                                  //{58, 5}, // Stellprogramm
				};

//				ClientDavInterface daf = this.getClientDavInterface();
//				DataModel config =daf.getDataModel();
//				AttributeGroup attributeGroup = config.getAttributeGroup("atg.tlsGloDeBlockAbruf");
//				Aspect aspect = config.getAspect("asp.tlsAbruf");
//				DataDescription dataDescriptionAbfrage = new DataDescription(attributeGroup, aspect);
//				Data dataAbfrage = daf.createData(attributeGroup);
//
//				for (int[] fehlendeDaten : missingData) {
//					dataAbfrage.getUnscaledValue("JobNummer").set(27);// TBD: wird nicht ausgewertet, DatKat anpassen
//					dataAbfrage.getUnscaledValue("Typ").set(fehlendeDaten[0]);// Typ
//					dataAbfrage.getUnscaledValue("ID").set(fehlendeDaten[1] + 16);// Bei Abruf ID + 16
//					ResultData result = new ResultData(_de.getDe(), dataDescriptionAbfrage, System.currentTimeMillis(), dataAbfrage);
//					this.sendData(result);
//				}

			}
			else {
				List deList = new ArrayList();
				deList.add(_de);
//				TlsApplicationLayer.writeNoDataToUnreachableDe(deList, _eakConversionModul.getTimeStamp(), -1, "atg.tlsGloDeFehler");
			}
		}
		return resultData;
	}

	/**
	 * FG 4 : Typ: 2 Id: 1 Name: Nicht darstellbare WVZ (Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                                 Erläuterung
	 * Byte 1           Länge DE-Block                              Länge des folgenden DE-Blocks [3 + n ]
	 * Byte 2           Daten-Endgeräte-Kanal                       [ 1.. 254 ]
	 * Byte 3           Typ der DE-Daten                            [ 2 ]
	 * Byte 4           Anzahl Code-Bytes                           Anzahl nicht mehr darstellbarer WVZ
	 * Byte 5           1. WVZ-Code                                 1. nicht mehr darstellb. Zeichen (lt. Anh. 7, Kap.1)
	 * Byte .. n.       WVZ-Code                                    n. nicht mehr darstellb. Zeichen (lt. Anh. 7, Kap. 1)
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ002Id001AtgTlsWzgNichtDarstellbareWVZAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// WVZ-Codes aus Telegramm -----------------------------------------------------------------------
		int count = deBlock.getDeData().readUnsignedByteAsInt();
		int[] _WvzCode_ = new int[count];
		for (int i = 0; i < count; i++) {
			_WvzCode_[i] = deBlock.getDeData().readUnsignedByteAsInt();
		}

		Data.NumberArray _NichtDarstellbareWVZ_ = data.getUnscaledArray("NichtDarstellbareWVZ");
		_NichtDarstellbareWVZ_.setLength(count);
		for (int i = 0; i < count; i++) {
			_NichtDarstellbareWVZ_.getValue(i).set(_WvzCode_[i]);
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 3 Id: 1 Name: Defekte Lampen (Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                               Erläuterung
	 * Byte 1           Länge DE-Block                            Länge des folgenden DE-Blocks [3 + n ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)                [ 1..254 ]
	 * Byte 3           Typ der DE-Daten                          [ 3 ]
	 * Byte 4           Anzahl Lampenpaare
	 * Byte 5           1. Lampen-Statusbyte                      Enthält die Gut/Schlecht-Aussage von 8 Lampen
	 * Byte .. n.       Lampen-Statusbyte                         Enthält die Gut/Schlecht-Aussage von 8 Lampen
	 * </pre>
	 * Inhalt 1. Lampen-Statusbyte
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *    NL 4     HL 4     NL 3     HL 3     NL 2     HL 2     NL 1     HL 1
	 * </pre>
	 * Bits 0,2,4,6 :    Hauptlampen 1 bis 4,    (0 = OK, 1 = defekt)<br> Bits 1,3,5,7 :    Nebenlampen 1 bis 4,    (0 =
	 * OK, 1 = defekt)<br> <br> Die weiteren Lampen-Statusbytes enthalten die Lampenpaare in entsprechend aufsteigender
	 * Folge.<br>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ003Id001AtgTlsWzgDefekteLampenAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Lampen-Paare (je 2 Lampen)  aus Telegramm -----------------------------------------------------------------------
		int countLightPairs = deBlock.getDeData().readUnsignedByteAsInt();

		int count;
		if ((countLightPairs % 4) == 0) {
			count = countLightPairs / 4;
		}
		else {
			count = (countLightPairs / 4) + 1;
		}
		int[] _LampenStatus_ = new int[count];
		for (int i = 0; i < count; i++) {
			_LampenStatus_[i] = deBlock.getDeData().readUnsignedByteAsInt();
		}

		Data.NumberArray _ZustandLampenPaar_ = data.getUnscaledArray("ZustandLampenPaar");
		_ZustandLampenPaar_.setLength(countLightPairs);
		int lightState;
		for (int i = 0; i < countLightPairs; i++) {
			lightState = (_LampenStatus_[i / 4] >>> (i % 4) * 2) & 0x03;
			_ZustandLampenPaar_.getValue(i).set(lightState);
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 17 Id: 80 Name: Betriebsart (Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                                    Erläuterung
	 * Byte 1           Länge DE-Block                                 Länge des folgenden DE-Blocks [ 3 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)                     [193..222, 223, (255)1]
	 * Byte 3           Typ der DE-Daten                               [ 17 ]
	 * Byte 4           Betriebsart                                    siehe unten
	 * </pre>
	 * 1) siehe Seite A6.1.1 - 3, Kapitel 1.4<br> Inhalt Byte: Betriebsart:<br>
	 * <pre>
	 *        0: frei
	 *        1: Normalbetrieb
	 *        2: Blindbetrieb
	 *        3: Handbetrieb
	 *        4: Autarker Betrieb
	 *        5: Testbetrieb
	 *        6: Notbetrieb
	 *   7..255: noch nicht definiert
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ017Id002AtgTlsWzgBetriebsArtAspParameterIst(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Betriebsart aus Telegramm -----------------------------------------------------------------------
		int _BetriebsArt_ = deBlock.getDeData().readUnsignedByteAsInt();

		data.getUnscaledValue("BetriebsArt").set(_BetriebsArt_);

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 17 Id: 2 Name: Betriebsart (Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                                    Erläuterung
	 * Byte 1           Länge DE-Block                                 Länge des folgenden DE-Blocks [ 3 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)                     [193..222, 223, (255)1]
	 * Byte 3           Typ der DE-Daten                               [ 17 ]
	 * Byte 4           Betriebsart                                    siehe unten
	 * </pre>
	 * 1) siehe Seite A6.1.1 - 3, Kapitel 1.4<br> Inhalt Byte: Betriebsart:<br>
	 * <pre>
	 *        0: frei
	 *        1: Normalbetrieb
	 *        2: Blindbetrieb
	 *        3: Handbetrieb
	 *        4: Autarker Betrieb
	 *        5: Testbetrieb
	 *        6: Notbetrieb
	 *   7..255: noch nicht definiert
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ017Id080AtgTlsWzgBetriebsArtAspTlsVorgabeBedienung(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Betriebsart aus Telegramm -----------------------------------------------------------------------
		int _BetriebsArt_ = deBlock.getDeData().readUnsignedByteAsInt();

		switch(_BetriebsArt_){
			case 1:   data.getTextValue("BetriebsArt").setText("Handbetrieb"); break;
			case 3:   data.getTextValue("BetriebsArt").setText("Normalbetrieb"); break;
			default: data.getUnscaledValue("BetriebsArt").set(_BetriebsArt_);
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 32 Id: 3 Name: WVZ-Grundeinstellung (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                                    Erläuterung
	 * Byte 1           Länge DE-Block                                 Länge des
	 * folgenden DE-Blocks [ 4 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)                     [ 1..254,
	 * 255 ]
	 * Byte 3           Typ der DE-Daten                               [ 32 ]
	 * Byte 4           Stellzustand: WVZ-Code
	 * anzuzeigendes WVZ
	 * Byte 5           Funktionsbyte                                   Ein / Aus /
	 * Blinken
	 * </pre>
	 * Inhalt Byte: Stellzustand:<br> Codierung von Verkehrszeichen: (WVZ-Code): Siehe Anhang 7, Kapitel 1.<br> Wird der
	 * WVZ-Code 255 ausgewählt, so bleibt der zuletzt geschaltete Stellzustand im Autarkbetrieb bestehen.<br> <br> Inhalt
	 * Funktionsbyte:<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *               Blinkzeit                            Funktion
	 * Bit 0..3:     Funktion:
	 *               0 = ausschalten
	 *               1 = einschalten
	 *               2 = blinken (Die Einstellung "WVZ blinkend" ist eine optionale
	 * Eigenschaft.)
	 *               3-15 reserviert
	 *
	 * Bit 4..7:     Blinkzeit in 200 ms (1..3 s)
	 *               Die Blinkzeit ist die gesamte Periodendauer, d.h. die Summe
	 * aus Ein- und Auszeit. Das Verhältnis Ein-
	 *               /Auszeit wird nach Vorgabe des Anlagenbetreibers am WZG
	 * eingestellt.
	 *               Die Koordination der Blinkzeiten verschiedener WVZ ist Aufgabe
	 * der Zentrale.
	 * </pre>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg004Typ032Id003AtgTlsWzgWvzGrundEinstellungAspParameterSoll(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		int _Stellzustand_;
		int _Funktion_ = 0;

		// Umsetzung
		Data data = result.getData();// Datensatz holen
		_Stellzustand_ = data.getItem("WVZAnzeige").getUnscaledValue("Code").intValue();

		// Wenn _Stellzustand_ = -1, muss Funktionsbyte auf "aus" gesetzt werden ( = 0).
		if (_Stellzustand_ == -1) {
			_Funktion_ = 0;// ausschalten
			_Stellzustand_ = 0;
		}
		else {
			if (data.getItem("WVZAnzeige").getItem("Blinken").getTextValue("Status").getText().equals("aus")) {
				_Funktion_ = 1;// einschalten ohne Blinken
			}
			else {
				_Funktion_ = 2;// einschalten mit Blinken

				// Blinkezeit auswerten.
				if (data.getItem("WVZAnzeige").getItem("Blinken").getTextValue("Dauer").getText().equals("aus")) {
					_Funktion_ &= 0x0f;// Bit 4-7 = 0 setzten
				}
				else {
					_Funktion_ |= data.getItem("WVZAnzeige").getItem("Blinken").getUnscaledValue("Dauer").intValue() << 4;
				}
			}
		}

		// DeBlock als TlsData aufbauen...

		TlsData deApplicationData = new TlsData(2);
		deApplicationData.writeIntAsByte(_Stellzustand_);
		deApplicationData.writeIntAsByte(_Funktion_);

		// DeBlock erzeugen
		try {
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 32, 3, _de.getDeChannel());
			debug.finest("TLS-DE-Block: " + HexDumper.toString(deBlock.getTlsDeDataBlock()));
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;
	}

	/**
	 * Entspricht {@link de.bsvrz.kex.tls.osi7.conversion.Fg004Default#Fg004Typ032Id003AtgTlsWzgWvzGrundEinstellungAspParameterIst(DeBlock,
	 *Data,DataDescription)}, jedoch werden die Daten als nachgeliefert gekennzeichnet.
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ032Id035AtgTlsWzgWvzGrundEinstellungAspParameterIst(DeBlock deBlock, Data data, DataDescription dataDescription) {
		// Generierte Methode: Standardverarbeitung wie bei Id 3, aber als "nachgelieferte" Daten an den DaV übergeben.
		ResultData resultData = Fg004Typ032Id003AtgTlsWzgWvzGrundEinstellungAspParameterIst(deBlock, data, dataDescription);
		return new ResultData(_de.getDe(), dataDescription, resultData.getDataTime(), resultData.getData(), true);
	}

	/**
	 * FG 4 : Typ: 32 Id: 3 Name: WVZ-Grundeinstellung (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                                    Erläuterung
	 * Byte 1           Länge DE-Block                                 Länge des
	 * folgenden DE-Blocks [ 4 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)                     [ 1..254,
	 * 255 ]
	 * Byte 3           Typ der DE-Daten                               [ 32 ]
	 * Byte 4           Stellzustand: WVZ-Code
	 * anzuzeigendes WVZ
	 * Byte 5           Funktionsbyte                                   Ein / Aus /
	 * Blinken
	 * </pre>
	 * Inhalt Byte: Stellzustand:<br> Codierung von Verkehrszeichen: (WVZ-Code): Siehe Anhang 7, Kapitel 1.<br> Wird der
	 * WVZ-Code 255 ausgewählt, so bleibt der zuletzt geschaltete Stellzustand im Autarkbetrieb bestehen.<br> <br> Inhalt
	 * Funktionsbyte:<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *               Blinkzeit                            Funktion
	 * Bit 0..3:     Funktion:
	 *               0 = ausschalten
	 *               1 = einschalten
	 *               2 = blinken (Die Einstellung "WVZ blinkend" ist eine optionale
	 * Eigenschaft.)
	 *               3-15 reserviert
	 *
	 * Bit 4..7:     Blinkzeit in 200 ms (1..3 s)
	 *               Die Blinkzeit ist die gesamte Periodendauer, d.h. die Summe
	 * aus Ein- und Auszeit. Das Verhältnis Ein-
	 *               /Auszeit wird nach Vorgabe des Anlagenbetreibers am WZG
	 * eingestellt.
	 *               Die Koordination der Blinkzeiten verschiedener WVZ ist Aufgabe
	 * der Zentrale.
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ032Id003AtgTlsWzgWvzGrundEinstellungAspParameterIst(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Betriebsart aus Telegramm -----------------------------------------------------------------------
		int _Stellzustand_ = deBlock.getDeData().readUnsignedByteAsInt();
		_validWvzCode = _Stellzustand_;
		int _Funktion_ = deBlock.getDeData().readUnsignedByteAsInt();

		// Wenn Funktion = ausgeschaltet, hat WVZ-Code keine Bedeutung. Wird deshalb für
		// Protokollausgaben auf -1 = "aus" gesetzt.
		if ((_Funktion_ & 0x0f) == 0) {
			_Stellzustand_ = -1;
		}
		data.getItem("WVZAnzeige").getUnscaledValue("Code").set(_Stellzustand_);

		if ((_Funktion_ & 0x02) == 0) {
			data.getItem("WVZAnzeige").getItem("Blinken").getUnscaledValue("Status").setText("aus");
			data.getItem("WVZAnzeige").getItem("Blinken").getUnscaledValue("Dauer").setText("aus");
		}
		else {
			data.getItem("WVZAnzeige").getItem("Blinken").getUnscaledValue("Status").setText("ein");
			int _BlinkDauer_ = _Funktion_ >>> 4;
			if ((_BlinkDauer_ < 5) || (_BlinkDauer_ > 15)) {
				debug.warning("Blinkdauer ungültig (5-15) : " + (_Funktion_ >>> 4) + ". Blinkdauerwert wurde auf 5 gesetzt!");
				_BlinkDauer_ = 5;
			}
			data.getItem("WVZAnzeige").getItem("Blinken").getUnscaledValue("Dauer").set(_BlinkDauer_);
		}

		data.getItem("WVZAnzeige").getItem("Status").getUnscaledValue("Fehler").setText("ok");
		data.getItem("WVZAnzeige").getItem("Status").getUnscaledValue("Programm").setText("abgeschlossen");

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 42 Id: 3 Name: Grundprogramm (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          [ 3 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)              [ 193..222, 223, (255), (evtl. 1..254,255)1 ]
	 * Byte 3           Typ der DE-Daten                        [ 42 ]
	 * Byte 4           Grundprogramm-Nummer                    [ 1 - 254, 255 ]
	 * 1) es wird der Clusterkanal gemäß der Definition Seite A6.1.1  3, Kapitel 1.4 angewendet, lediglich bei Anlagen,
	 * die
	 * nach dem Prinzip der alten FG5 (TLS Ausgabe 1993) arbeiten, können die E/A-Kanäle direkt adressiert werden
	 * </pre>
	 * Inhalt Byte Programm-Nummer:<br> Nummer des Programms, in das bei Autarkbetrieb umgeschaltet werden soll. Wird die
	 * Programmnummer 255 ausgewählt, so bleibt das vor Beginn des Autarkbetrieb geschaltete Programm im Autarkbetrieb
	 * bestehen.<br>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ042Id003AtgTlsWzgGrundProgrammAspParameterIst(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Grundprogramm aus Telegramm --------------------------------------------------------------------------------
		int _GrundProgramm_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("GrundProgramm").set(_GrundProgramm_);

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 48 Id: 5 und Id 37 Name: WVZ-Stellzustand (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden
	 * DE-Blocks [ 4 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)              [ 1..254, 255 ]
	 * Byte 3           Typ der DE-Daten                        [ 48 ]
	 * Byte 4           Stellzustand: Code                      anzuzeigendes WVZ
	 * Byte 5           Funktionsbyte                           Ein / Aus /
	 * Blinken
	 * </pre>
	 * Inhalt Byte Stellzustand:<br> Codierung von Verkehrszeichen: (WVZ-Code): Siehe Anhang 7, Kapitel 1.<br> <br> Inhalt
	 * Funktionsbyte:
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *               Blinkzeit                            Funktion
	 * Bit 0..3:     Funktion:
	 *               0 = ausschalten
	 *               1 = einschalten
	 *               2 = blinken (Die Einstellung "WVZ blinkend" ist eine optionale
	 * Eigenschaft.)
	 *               3-15 reserviert
	 *
	 * Bit 4..7:     Blinkzeit in 200 ms (1..3 s)
	 *               Die Blinkzeit ist die gesamte Periodendauer, d.h. die Summe
	 * aus Ein- und Auszeit. Das Verhältnis Ein-
	 *               /Auszeit wird nach Vorgabe des Anlagenbetreibers am WZG
	 * eingestellt.
	 *               Eine sinnvolle Abstimmung der Periodendauern für verschiedene
	 * WVZ ist Aufgabe der Zentrale.
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ048Id005AtgTlsWzgWvzStellZustandAspTlsZustand(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Betriebsart aus Telegramm -----------------------------------------------------------------------
		int _Stellzustand_ = deBlock.getDeData().readUnsignedByteAsInt();
		int _Funktion_ = deBlock.getDeData().readUnsignedByteAsInt();

		// Wenn Funktion = ausgeschaltet, hat WVZ-Code keine Bedeutung. Wird deshalb für
		// Protokollausgaben auf -1 = "aus" gesetzt.
		if ((_Funktion_ & 0x0f) == 0) {
			_Stellzustand_ = -1;
		}
		data.getUnscaledValue("WvzCode").set(_Stellzustand_);

		if ((_Funktion_ & 0x02) == 0) {
			data.getItem("Blinken").getUnscaledValue("Status").setText("aus");
			data.getItem("Blinken").getUnscaledValue("Dauer").setText("aus");
		}
		else {
			data.getItem("Blinken").getUnscaledValue("Status").setText("ein");
			int _BlinkDauer_ = _Funktion_ >>> 4;
			if ((_BlinkDauer_ < 5) || (_BlinkDauer_ > 15)) {
				debug.warning("Blinkdauer ungültig (5-15) : " + (_Funktion_ >>> 4) + ". Blinkdauerwert wurde auf 5 gesetzt!");
				_BlinkDauer_ = 5;
				return resultData;
			}
			data.getItem("Blinken").getUnscaledValue("Dauer").set(_BlinkDauer_);
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 48 Id: 5 und Id 37 Name: WVZ-Stellzustand (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden
	 * DE-Blocks [ 4 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)              [ 1..254, 255 ]
	 * Byte 3           Typ der DE-Daten                        [ 48 ]
	 * Byte 4           Stellzustand: Code                      anzuzeigendes WVZ
	 * Byte 5           Funktionsbyte                           Ein / Aus /
	 * Blinken
	 * </pre>
	 * Inhalt Byte Stellzustand:<br> Codierung von Verkehrszeichen: (WVZ-Code): Siehe Anhang 7, Kapitel 1.<br> <br> Inhalt
	 * Funktionsbyte:
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *               Blinkzeit                            Funktion
	 * Bit 0..3:     Funktion:
	 *               0 = ausschalten
	 *               1 = einschalten
	 *               2 = blinken (Die Einstellung "WVZ blinkend" ist eine optionale
	 * Eigenschaft.)
	 *               3-15 reserviert
	 *
	 * Bit 4..7:     Blinkzeit in 200 ms (1..3 s)
	 *               Die Blinkzeit ist die gesamte Periodendauer, d.h. die Summe
	 * aus Ein- und Auszeit. Das Verhältnis Ein-
	 *               /Auszeit wird nach Vorgabe des Anlagenbetreibers am WZG
	 * eingestellt.
	 *               Eine sinnvolle Abstimmung der Periodendauern für verschiedene
	 * WVZ ist Aufgabe der Zentrale.
	 * </pre>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg004Typ048Id081AtgTlsWzgWvzStellZustandAspTlsVorgabe(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		int _Stellzustand_;
		int _Funktion_ = 0;

		// Umsetzung
		Data data = result.getData();// Datensatz holen
		_Stellzustand_ = data.getUnscaledValue("WvzCode").intValue();

		// Wenn _Stellzustand_ = -1, muss Funktionsbyte auf "aus" gesetzt werden ( = 0).
		if (_Stellzustand_ == -1) {
			_Funktion_ = 0;// ausschalten
			_Stellzustand_ = 0;
		}
		else {
			if (data.getItem("Blinken").getTextValue("Status").getText().equals("aus")) {
				_Funktion_ = 1;// einschalten ohne Blinken
			}
			else {
				_Funktion_ = 2;// einschalten mit Blinken

				// Blinkezeit auswerten.
				if (data.getItem("Blinken").getTextValue("Dauer").getText().equals("aus")) {
					_Funktion_ &= 0x0f;// Bit 4-7 = 0 setzten
				}
				else {
					_Funktion_ |= data.getItem("Blinken").getUnscaledValue("Dauer").intValue() << 4;
				}
			}
		}

		// DeBlock als TlsData aufbauen...

		TlsData deApplicationData = new TlsData(2);
		deApplicationData.writeIntAsByte(_Stellzustand_);
		deApplicationData.writeIntAsByte(_Funktion_);

		// DeBlock erzeugen
		try {
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 48, 81, _de.getDeChannel());
			debug.finest("TLS-DE-Block: " + HexDumper.toString(deBlock.getTlsDeDataBlock()));
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;
	}

	/**
	 * FG 4 : Typ: 49 Id: 5  Name: Helligkeit (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                                    Erläuterung
	 * Byte 1           Länge DE-Block                                 Länge des folgenden DE-Blocks [ 4 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)                     [193..222, 223, (255)1 ]
	 * Byte 3           Typ der DE-Daten                               [ 49 ]
	 * Byte 4           Helligkeitswert                                [ 0..100 Prozent ]
	 * Byte 5           Helligkeitsstatusbyte                          siehe unten
	 * 1) siehe Anhang 6, Teil 1, Kapitel 1.4
	 * </pre>
	 * Inhalt Byte: Helligkeitswert:<br> Der Helligkeitswert für WZGs wird unabhängig von dem Anzeigeprinzip des WZG als
	 * eine Prozentzahl von 0 bis 100 übertragen. WZGs stellen die dem Prozentwert am nächsten liegende Helligkeitsstufe
	 * ein. Die hierbei geltenden lichttechnischen Anforderungen werden in den "vorläufigen lichttechnischen Festlegungen
	 * für WVZ und Fahrstreifensignalgeber nach dem Lichtrasterprinzip" der BASt beschrieben.<br> In Nachrichten in
	 * Abrufrichtung, die im Helligkeitsstatusbyte die automatische Helligkeitsteuerung einschalten, ist der
	 * Helligkeitswert nicht definiert.<br> <br> Inhalt Helligkeitsstatusbyte:<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *    X        X        X        X        X        X        Meld.    Autom.
	 *
	 * Bit 0:   Automatik:
	 *          0 = Helligkeitswert wird durch die Zentrale ferneingestellt
	 *          1 = automatische, lokale Helligkeitssteuerung (optional, s.o.)
	 * Bit 1:   Meldungen im Automatikbetrieb:
	 *          0 = keine spontanen Meldungen des aktuellen Helligkeitswertes
	 *          1 = nach jeder Änderung des eingestellten Helligkeitswertes spontane Meldung senden
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ049Id005AtgTlsWzgHelligkeitAspTlsZustand(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Helligkeitswert aus Telegramm -------------------------------------------------------------------------------
		int _Helligkeitswert_ = deBlock.getDeData().readUnsignedByteAsInt();

		if ((_Helligkeitswert_ >= 0) && (_Helligkeitswert_ <= 100)) {
			data.getUnscaledValue("HelligkeitsWert").set(_Helligkeitswert_);
		}
		else {
			debug.warning("Helligkeitswert ungültig : " + _Helligkeitswert_ + Debug.NEWLINE +
			              "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE +
			              deBlock.toString());
			return resultData;
		}
		// Helligkeitswert aus Telegramm -------------------------------------------------------------------------------
		int _Helligkeitsstatusbyte_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("Status").set(_Helligkeitsstatusbyte_ & 0x03);

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 50 Id: 5 und Id 37 Name: Wechseltext (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden
	 * DE-Blocks [ 4 + Textl. ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)              [ 1..254, 255 ]
	 * Byte 3           Typ der DE-Daten                        [ 50 ]
	 * Byte 4           Funktionsbyte                           Ein / Aus /
	 * Blinken
	 * Byte 5           Anzahl Textzeichen                      [ 0..220 ]
	 * Byte 6           1. Textzeichen in ASCII-Code            angezeigter /
	 * anzuzeigender Text
	 * Byte 7           2. Textzeichen in ASCII-Code            (siehe Anhang 7,
	 * Kap.3)
	 * ......
	 * Byte ..          n. Textzeichen in ASCII-Code
	 * </pre>
	 * Inhalt Funktionsbyte:<br> Codierung von Verkehrszeichen: (WVZ-Code): Siehe Anhang 7, Kapitel 1.<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *               Blinkzeit                            Funktion
	 * Bit 0..3:     Funktion:
	 *               0 = ausschalten
	 *               1 = einschalten
	 *               2 = blinken (Die Einstellung "WVZ blinkend" ist eine optionale
	 * Eigenschaft.)
	 *               3-15 reserviert
	 *
	 * Bit 4..7:     Blinkzeit in 200 ms (1..3 s)
	 *               Die Blinkzeit ist die gesamte Periodendauer, d.h. die Summe
	 * aus Ein- und Auszeit. Das Verhältnis Ein-
	 *               /Auszeit wird nach Vorgabe des Anlagenbetreibers am WZG
	 * eingestellt.
	 *               Eine sinnvolle Abstimmung der Periodendauern für verschiedene
	 * WVZ ist Aufgabe der Zentrale.
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ050Id005AtgTlsWzgWechselTextAspTlsZustand(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Funktionsbyte aus Telegramm ---------------------------------------------------------------------------------
		int _Funktion_ = deBlock.getDeData().readUnsignedByteAsInt();

		if ((_Funktion_ & 0x02) == 0) {
			data.getItem("Blinken").getUnscaledValue("Status").setText("aus");
			data.getItem("Blinken").getUnscaledValue("Dauer").setText("aus");
		}
		else {
			data.getItem("Blinken").getUnscaledValue("Status").setText("ein");
			int _BlinkDauer_ = _Funktion_ >>> 4;
			if ((_BlinkDauer_ < 5) || (_BlinkDauer_ > 15)) {
				debug.warning("Blinkdauer ungültig (5-15) : " + (_Funktion_ >>> 4) + ". Blinkdauerwert wurde auf 5 gesetzt!");
				_BlinkDauer_ = 5;
			}
			data.getItem("Blinken").getUnscaledValue("Dauer").set(_BlinkDauer_);
		}

		// Wenn Funktion = ausgeschaltet, hat WVZ-Code keine Bedeutung. Wird deshalb für
		// Protokollausgaben auf leeren String gesetzt.

		data.getTextValue("WVZTextZeichen").setText("");

		if ((_Funktion_ & 0x0f) != 0) {
			int count = deBlock.getDeData().readUnsignedByteAsInt();

			StringBuffer _AnzeigeText_ = new StringBuffer();
			for (int i = 0; i < count; i++) {
				_AnzeigeText_.append(String.format("%c", deBlock.getDeData().readUnsignedByteAsInt()));
			}
			data.getTextValue("WVZTextZeichen").setText(_AnzeigeText_.toString());
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 4 : Typ: 50 Id: 5 und Id 37 Name: Wechseltext (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden
	 * DE-Blocks [ 4 + Textl. ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)              [ 1..254, 255 ]
	 * Byte 3           Typ der DE-Daten                        [ 50 ]
	 * Byte 4           Funktionsbyte                           Ein / Aus /
	 * Blinken
	 * Byte 5           Anzahl Textzeichen                      [ 0..220 ]
	 * Byte 6           1. Textzeichen in ASCII-Code            angezeigter /
	 * anzuzeigender Text
	 * Byte 7           2. Textzeichen in ASCII-Code            (siehe Anhang 7,
	 * Kap.3)
	 * ......
	 * Byte ..          n. Textzeichen in ASCII-Code
	 * </pre>
	 * Inhalt Funktionsbyte:<br> Codierung von Verkehrszeichen: (WVZ-Code): Siehe Anhang 7, Kapitel 1.<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *               Blinkzeit                            Funktion
	 * Bit 0..3:     Funktion:
	 *               0 = ausschalten
	 *               1 = einschalten
	 *               2 = blinken (Die Einstellung "WVZ blinkend" ist eine optionale
	 * Eigenschaft.)
	 *               3-15 reserviert
	 *
	 * Bit 4..7:     Blinkzeit in 200 ms (1..3 s)
	 *               Die Blinkzeit ist die gesamte Periodendauer, d.h. die Summe
	 * aus Ein- und Auszeit. Das Verhältnis Ein-
	 *               /Auszeit wird nach Vorgabe des Anlagenbetreibers am WZG
	 * eingestellt.
	 *               Eine sinnvolle Abstimmung der Periodendauern für verschiedene
	 * WVZ ist Aufgabe der Zentrale.
	 * </pre>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg004Typ050Id081AtgTlsWzgWechselTextAspTlsVorgabe(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		int _Funktion_ = 0;

		// Umsetzung
		Data data = result.getData();// Datensatz holen
		int count = data.getTextValue("WVZTextZeichen").getText().length();
		String text = data.getTextValue("WVZTextZeichen").getText();

		// Wenn Länge = 0, muss Funktionsbyte auf "aus" gesetzt werden
		if (count == 0) {
			_Funktion_ = 0;// ausschalten
		}
		else {
			if (data.getItem("Blinken").getTextValue("Status").getText().equals("aus")) {
				_Funktion_ = 1;// einschalten ohne Blinken
			}
			else {
				_Funktion_ = 2;// einschalten mit Blinken

				// Blinkezeit auswerten.
				if (data.getItem("Blinken").getTextValue("Dauer").getText().equals("aus")) {
					_Funktion_ &= 0x0f;// Bit 4-7 = 0 setzten
				}
				else {
					_Funktion_ |= data.getItem("Blinken").getUnscaledValue("Dauer").intValue() << 4;
				}
			}
		}

		// DeBlock als TlsData aufbauen...

		TlsData deApplicationData = new TlsData(1 + count);
		deApplicationData.writeIntAsByte(_Funktion_);
		deApplicationData.writeIntAsByte(count);

		deApplicationData.writeByteArray(text.getBytes());

		// DeBlock erzeugen
		try {
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 50, 81, _de.getDeChannel());
			debug.finest("TLS-DE-Block: " + HexDumper.toString(deBlock.getTlsDeDataBlock()));
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;
	}

	/**
	 * FG 4 : Typ: 55 Id: 5 Name: Stellzustand (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                                    Erläuterung
	 * Byte 1           Länge DE-Block                                 [5 | 6+f*n]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)                     [ 1..254 ]
	 * Byte 3           Typ der DE-Daten                               [ 55 ]
	 * Byte 4           Anzeigeprinzip                                 siehe unten
	 * Byte 5           Stellcode                                      anzuzeigendes WVZ/Programm
	 * Byte 6           Funktionsbyte                                  siehe unten
	 * Byte 7           n : Anzahl Komponenten                         1 -[213 | 106] siehe unten
	 * f Bytes...       Stellzustand der 1.Komponente                  siehe unten
	 * f Bytes          Stellzustand der n.Komponente
	 * </pre>
	 * Inhalt Byte : Anzeigeprinzip<br> Das Byte gibt das Anzeigeprinzip des adressierten/sendenden DEs an und ermöglicht
	 * damit die genaue Interpretation des folgenden Telegramminhalts. Es ist folgendermaßen codiert:<br>
	 * <pre>
	 * Anzeigeprinzip        Wert
	 *              a        0
	 *              b        1
	 *              c        2
	 *              d        3
	 *              e        4
	 *        Cluster        8
	 * </pre>
	 * Erhält ein DE in Abrufrichtung eine Nachricht mit einem Wert, der nicht seinem Anzeigeprinzip entspricht, so wird
	 * die Nachricht mit Code 21 falsches Anzeigeprinzip negativ quittiert.<br> <br> Inhalt Byte: Stellcode<br> Der
	 * Inhalt des Bytes ist abhängig davon, ob das adressierte DE ein Clusterkanal ist bzw. vom Anzeigeprinzip des
	 * DEs:<br>
	 * <pre>
	 * Clusterkanäle :   Nummer des Programms, das eingestellt werden soll. Ist der Wert 0, so behalten alle DEs
	 *                   den Zustand , den sie zuletzt hatten. In der Folge akzeptieren diese DEs individuell an sie
	 *                   gerichtete Stellbefehle und ändern ihren Zustand entsprechend. Ist als Wert eine
	 *                   Programmnummer zwischen 1 und 254 angegeben, so werden alle DEs entsprechend der
	 *                   Definition dieses Programms eingestellt. Ausnahmen hiervon bilden lediglich DEs, für
	 *                   welche die entsprechende Programmdefinition der Wert 255 "don´t care" enthält. Diese
	 *                   DEs behalten ihren Zustand und sind in der Folge durch individuelle Stellbefehle steuerbar.
	 *                   Der Wert 255 ist nicht zulässig und wird mit Code 43 negativ quittiert.
	 * Anzeigeprinzip a und b :   WVZ-Code entsprechend Anhang 7, Kapitel 1. Gibt den WVZ-Code an, den das
	 *                   DE darstellen soll. Die Nachricht wird nur akzeptiert, wenn der Stellcode des Clusterkanals
	 *                   den Wert 0 enthält oder die Definition des am Cluster eingestellten Programms für dieses
	 *                   DE den Wert 255 "don´t care" angibt. In allen anderen Fällen wird der Stellbefehl mit Code
	 *                   10 "Befehl bzw. Programm in dieser Betriebsart nicht ausführbar" negativ quittiert. Der
	 *                   Wert 0 ist für DEs dieser Anzeigeprinzipien nicht zulässig und wird ebenfalls negativ
	 *                   quittiert.
	 * Anzeigeprinzip c und d :   Stellcode des vordefinierten Textes, den das DE darstellen soll. Die Nachricht
	 *                   wird nur akzeptiert, wenn der eingestellte Stellcode des Clusterkanals den Wert 0 enthält
	 *                   oder die Definition des am Cluster eingestellten Programms für dieses DE den Wert 255
	 *                   "don´t care" angibt. In allen anderen Fällen wird der Stellbefehl mit Code 10 "Befehl bzw.
	 *                   Programm in dieser Betriebsart nicht ausführbar" negativ quittiert. Ist der Wert 0, so wird
	 *                   das Telegramm in der Langversion übertragen und enthält den darzustellenden Text als
	 *                   ASCII-String.
	 * Anzeigeprinzip e :   Stellcode der vordefinierten Gruppendefinition, die das DE darstellen soll. Die
	 *                   Nachricht wird nur akzeptiert, wenn der eingestellte Stellcode des Clusterkanals den Wert 0
	 *                   enthält oder die Definition des am Cluster eingestellten Programms für dieses DE den Wert
	 *                   255 "don´t care" angibt. In allen anderen Fällen wird der Stellbefehl mit Code 10 "Befehl
	 *                   bzw. Programm in dieser Betriebsart nicht ausführbar" negativ quittiert. Ist der Wert 0, so
	 *                   wird das Telegramm in der Langversion übertragen und enthält für jede Komponente den
	 *                   entsprechenden Stellzustand.
	 * </pre>
	 * Zulässige Werte:
	 * <pre>
	 *                       Cluster                      a,b                      c,d,e
	 * Steuerungsprinzip 1   0                            --                       0
	 * Steuerungsprinzip 2   1-254                        Anhang 7, Kap. 1         1-254
	 * Steuerungsprinzip 3   0,1-254                      --                       0,1-254
	 * </pre>
	 * Inhalt Funktionsbyte:<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *               Blinkzeit                      Status       Funktion
	 * </pre>
	 * Funktion
	 * <pre>
	 * Bit    Wert   Bedeutung                           Kommentar
	 * 0..1   0      ausschalten                         nur für DEs nach Anzeigeprinzip b und d
	 *        1      einschalten                         sowie Blinker nach Anzeigeprinzip a
	 *        2      blinken                             nur bei DEs nach Anzeigeprinzip b und d, optional
	 *        3      reserviert
	 * DEs, die in Abrufrichtung ein Funktionsbyte mit einem Wert erhalten, der gemäß ihrem Anzeigeprinzip
	 * nicht zulässig ist, antworten mit einer negativen Quittung, Code 14 "Funktionsbyte unzulässig".
	 * </pre>
	 * Status (nur in Antwortrichtung):
	 * <pre>
	 * Bit    Wert   Bedeutung                           Kommentar
	 * 2      0      kein Fehler aufgetreten
	 *        1      Fehler aufgetreten
	 * 3      2      Programm abgeschlossen              wird nur verwendet von Clusterkanälen und DEs, die
	 *        3      Programm noch nicht abgeschlossen   Programmschaltungen ausführen
	 * Blinkzeit<br>
	 *
	 * Bit    Wert   Bedeutung                           Kommentar
	 * 4..7   5-15   Blinkzeit in 200ms (1..3 s)         nur für DEs nach Anzeigeprinzip b und d
	 *        0                                          wenn Funktion ungleich "blinken"
	 * Die Blinkzeit ist die gesamte Periodendauer, d.h. die Summe aus Ein- und Auszeit. Das Verhältnis Ein-
	 * /Auszeit wird nach Vorgabe des Anlagenbetreibers am WZG eingestellt.
	 * Eine sinnvolle Abstimmung der Periodendauern für verschiedene WVZ ist Aufgabe der Zentrale.
	 * </pre>
	 * <pre>
	 * Inhalt Anzahl Komponenten :
	 * Wird die Nachricht in der Langversion versendet (nur bei Stellcode gleich 0), so gibt das Byte die Anzahl der
	 * Komponenten an, für die in der Folge Stellzustände übermittelt werden.
	 * Die Langversion ist nur für WZG nach Anzeigeprinzip c)-e) zulässig, für Anzeigen nach Prinzip a) und b) ist
	 * sie nicht anwendbar und für Clusterkanäle werden individuelle Stellbefehle an die jeweiligen DEs direkt
	 * adressiert.
	 * Inhalt : Stellzustand einer Komponente:
	 * Für WZG vom Anzeigeprinzip a) und b) (feste Zeicheninhalte) ist das Telegramm in der Langversion nicht
	 * zulässig, da es bedeutungslos wäre.
	 * Für WZG vom Anzeigeprinzip c) und d) (Wechseltext) besteht der Stellzustand einer Komponente aus jeweils
	 * nur einem Byte und ist wie folgt aufgebaut :
	 *
	 * Byte 1            Textzeichen in ASCII-Code            siehe Anhang 7, Kapitel 3
	 *
	 * Für WZG vom Anzeigeprinzip e) (Gruppen) besteht der Stellzustand einer Komponente aus jeweils 2 Bytes
	 * und ist wie folgt aufgebaut:<br>
	 *
	 * Byte 1            WVZ-Code                             siehe Anhang 7, Kapitel 1
	 * Byte 2            Funktionsbyte                        siehe oben
	 * </pre>
	 * Das Funktionsbyte ist definiert wie oben. Komponenten nach Anzeigeprinzip a) melden in Antwortrichtung
	 * den Sachverhalt Störung erkannt - Prisma nicht bewegbar / Blinker defekt mit gesetztem Bit 2 im
	 * Funktionsbyte. Kann der eingestellte WVZ-Code vom Gerät noch ermittelt werden, so wird er in Byte 1
	 * übertragen. Ansonsten wird der WVZ-Code 245 undefinierte Stellung gemeldet.
	 * Für Clusterkanäle ist die Nachricht in der Langversion nicht zulässig, da die Stellbefehle an die einzelnen DEs
	 * direkt adressiert werden können.<br>
	 * <br>
	 * Die folgende Tabelle stellt übersichtlich alle Varianten des Telegramms zusammen, erläutert deren Funktion und gibt
	 * an, für welche Kombination aus Anzeige und Steuerungsprinzip die jeweilige Verwendung zulässig ist. Dabei sind
	 * Felder -- gekennzeichnet, wenn sie im betreffenden Kontext nicht relevant sind. Mit - gekennzeichnete Felder in der
	 * Spalte "Steuerungsprinzip" bedeuten, dass die jeweiligen Stellbefehle negativ quittiert werden, die Nummer in
	 * Klammern gibt den zu verwendenden Fehlercode an. Felder mit zul kennzeichnen die Zulässigkeit des
	 * entsprechenden Telegramms für das jeweilige DE. Eine hinter zul angefügte Nummer  verweist auf
	 * einen Telegrammtyp aus früheren Versionen der TLS, die bei Altlagen im jeweiligen Kontext noch zulässig sind.<br>
	 * <pre>
	 * ID   Typ   Länge  Stellcode  n      Funktion                       Anzeige      f     Steuerungsprinzip
	 *                                                                    prinzip
	 *                                                                                       1        2       3
	 * 5R   55    5      1-254      --     Stellbefehl mit Codenummer     a            --    --       zul     --
	 *                              --                                    b            --    --       zul 48  --
	 *                              --                                    c,d,Cl       --    -(45)    zul     zul
	 *                              --                                    e            --    -(45)    zul 58  zul 58
	 *                   0          --     Stellbefehl mit freier         a-e          --    -(44)    -(43)   -(44)
	 *                              --     Definition                     Cl           --    zul      -(43)   zul
	 *            6+n*f  0          1 --   Stellbefehl mit freier         a,b          --    --       -(46)   --
	 *                              1 213  Definition                     c,d          1     zul 50   -(46)   zul 50
	 *                              1 106                                 e            2     zul      -(46)   zul
	 *                              1 --                                  Cl           --    -(46)    -(46)   -(46)
	 * 5A         5      1-254      --     Stellzustand mit Codenummer    a            --    --       zul     --
	 *                              --                                    b            --    --       zul 48  --
	 *                              --                                    c,d,Cl       --    --       zul     zul
	 *                              --                                    e            --    --       zul 58  zul 58
	 *                   0          --     Stellzustand mit freier        a-e          --    --       --      --
	 *                              --     Definition                     Cl           --    --       zul     zul
	 *            6+n*f  0          1 --   Stellzustand mit freier        a,b          --    --       --      --
	 *                              1 213  Definition                     c,d          1     zul 50   --      zul 50
	 *                              1 106                                 e            2     zul      --      zul
	 *                              1 --                                  Cl           --    --       --      --
	 * 21R        2      --         --     Abruf Stellzustand             alle         --    zul      zul     zul
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ055Id005AtgTlsWzgStellZustandAspTlsZustand(DeBlock deBlock, Data data, DataDescription dataDescription) {
		// Analog zur Grundeinstellung
		ResultData resultData = Fg004Typ033Id003AtgTlsWzgGrundEinstellungAspParameterIst(deBlock, data, dataDescription);

		//Ergebnisdatensatz erzeugen...
		return resultData;
	}

	/**
	 * FG 4 : Typ: 33 Id: 3 und Id: 31 Name: Grundeinstellung (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                              Erläuterung
	 * Byte 1           Länge DE-Block                           Länge des folgenden DE-Blocks [5 |6+f*n]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)               [ 1..254 ]
	 * Byte 3           Typ der DE-Daten                         [ 33 ]
	 * Byte 4           Anzeigeprinzip                           siehe unten
	 * Byte 5           Stellcode der Grundeinstellung           [0, 1..254, 255]siehe unten
	 * Byte 6           Funktionsbyte                            siehe unten, vgl Kap. 6.2.18
	 * Byte 7           n : Anzahl Komponenten 1             [213 | 106] siehe unten
	 * f Bytes...       Grundeinstellung der 1.Komponente        siehe unten
	 * f Bytes          Grundeinstellung der n.Komponente
	 * </pre>
	 * Inhalt Byte : Anzeigeprinzip<br> Das Byte gibt das Anzeigeprinzip des adressierten/sendenden DEs an und ermöglicht
	 * damit die genaue Interpretation des folgenden Telegramminhalts. Es ist folgendermaßen codiert:<br>
	 * <pre>
	 * Anzeigeprinzip      Wert
	 *              a      0
	 *              b      1
	 *              c      2
	 *              d      3
	 *              e      4
	 *        Cluster      8
	 * </pre>
	 * Erhält ein DE in Abrufrichtung eine Nachricht mit einem Wert, der nicht seinem Anzeigeprinzip entspricht, so wird
	 * die Nachricht mit Code 21 "falsches Anzeigeprinzip" negativ quittiert.<br> Inhalt Byte: Stellcode der
	 * Grundeinstellung:<br> Der Inhalt des Bytes ist abhängig davon, ob das adressierte DE ein Clusterkanal ist bzw. vom
	 * Anzeigeprinzip des DEs. Diese Information ist in Byte 4 "Anzeigeprinzip" enthalten.<br>
	 * <pre>
	 * Clusterkanäle :   Nummer des Programms, das im Autarkbetrieb eingenommen werden soll. Ist der Wert 0,
	 *                   so nehmen alle DEs dann den Zustand ein, der ihrer jeweiligen individuellen
	 *                   Grundeinstellung entspricht. Ist der Wert 255, so behalten alle DEs den zuletzt eingestellten
	 *                   Zustand bei. Ist als Wert eine Programmnummer zwischen 1 und 254 angegeben, so werden
	 *                   alle DEs im Autarkbetrieb entsprechend der Definition dieses Programms eingestellt. Die
	 *                   Codedefinition wird für alle DEs übernommen und von diesen spontan gemeldet.
	 *                   Ausnahmen hiervon bilden lediglich DEs, für welche die entsprechende
	 *                   Programmdefinition den Wert 255 don´t care enthält. Diese DEs nehmen im
	 *                   Autarkbetrieb den Stellcode ihrer individuellen Grundeinstellung an.
	 * Anzeigeprinzip a und b :  WVZ-Code entsprechend Anhang 7, Kapitel 1. Gibt den WVZ-Code an, den das
	 *                           DE im Autarkbetrieb darstellen soll, falls die Grundeinstellungs-Programmnummer des
	 *                           Clusterkanals den Wert 0 enthält oder die Definition des Grundeinstellungsprogramms für
	 *                           dieses DE den Wert 255 "don´t care" angibt. In allen anderen Fällen ist die individuelle
	 *                           Grundeinstellung nicht relevant. Ist der Wert 255, so behält das WZG seinen letzten
	 *                           Stellzustand bei. Der Wert 0 ist für DEs dieser Anzeigeprinzipien nicht zulässig.
	 * Anzeigeprinzip c und d :  Stellcode des vordefinierten Textes, den das DE im Autarkbetrieb darstellen soll,
	 *                           falls die Grundeinstellungs-Programmnummer des Clusterkanals den Wert 0 enthält oder
	 *                           die Definition des Grundeinstellungsprogramms für dieses DE den Wert 255 "don´t care"
	 *                           angibt. In allen anderen Fällen ist die individuelle Grundeinstellung nicht relevant. Ist
	 * der
	 *                           Wert 255, so behält das WZG seinen zuletzt dargestellten Text bei. Ist der Wert hingegen
	 *                           0, so wird das Telegramm in der Langversion übertragen und enthält den darzustellenden
	 *                           Text als ASCII-String.
	 * Anzeigeprinzip e :        Stellcode der vordefinierten Gruppendefinition, die das DE im Autarkbetrieb
	 *                           darstellen soll, falls die Grundeinstellungs-Programmnummer des Clusterkanals den Wert 0
	 *                           enthält oder die Definition des Grundeinstellungsprogramms für dieses DE den Wert 255
	 *                           "don´t care" angibt. In allen anderen Fällen ist die individuelle Grundeinstellung nicht
	 *                           relevant. Ist der Wert 255, so behält die Gruppe ihren zuletzt dargestellten Inhalt bei.
	 * Ist der
	 *                           Wert hingegen 0, so wird das Telegramm in der Langversion übertragen und enthält für
	 *                           jede Komponente den entsprechenden Stellzustand.
	 * </pre>
	 * Zulässige Werte :
	 * <pre>
	 *                            Cluster                a,b               c,d,e
	 * Steuerungsprinzip 1        0, 255                 --                0, 255
	 * Steuerungsprinzip 2        1-254, 255             Anhang 7, Kap. 1  1-254,255
	 * Steuerungsprinzip 3        0,1-254,255            --                0,1-254,255
	 * </pre>
	 * Inhalt Funktionsbyte :<br> Der Inhalt des Funktionsbytes entspricht der verallgemeinerten Definition des
	 * Funktionsbytes in Kapitel 6.2.18.<br> <br> Inhalt Grundeinstellung einer Komponenten :<br> Dieser Teil des
	 * Telegramms wird genauso wie die Anzahl der Komponenten im Byte 7 nur in der Langversion des Telegramms übertragen,
	 * d.h. insbesondere, dass der Stellcode in Byte 5 den Wert 0 haben muss. Wird die Langversion mit einem Stellcode
	 * ungleich 0 empfangen, so wird mit Code 46 Langversion nicht zulässig negativ quittiert. Inhalt und Länge der
	 * wiederholten Komponentenblöcke sind abhängig vom Anzeigeprinzip des adressierten DE:<br>
	 * <pre>
	 * Clusterkanäle : Die Langversion ist für Clusterkanäle nicht zulässig. Die individuellen Grundeinstellungen
	 *                 werden an die einzelnen DEs direkt adressiert. Negative Quittung Code 46.
	 * Anzeigeprinzip a und b : Die Langversion ist inhaltsgemäß für diese Anzeigen nicht zulässig. Negative
	 *                 Quittung Code 46.
	 * Anzeigeprinzip c und d : Die Länge eines Komponentenblocks (f) ist 1, so dass die Langversion eine DEBlock-
	 *                 Länge von 6 + n hat. Die Anzahl Komponenten (Textzeichen) ist deswegen auf
	 *                 maximal 213 beschränkt. Jeder Komponentenblock ist folgendermaßen aufgebaut:
	 *
	 *                 Byte 1           Textzeichen in ASCII-Code        siehe Anhang 7, Kapitel 3
	 *
	 * Anzeigeprinzip e : Die Länge eines Komponentenblocks (f) ist 2, so dass die Langversion eine DE-Block-
	 *                 Länge von 6 + 2*n hat. Die Anzahl Komponenten ist deswegen auf maximal 106
	 *                 beschränkt. Jeder Komponentenblock ist folgendermaßen aufgebaut:
	 *
	 *                 Byte 1           WVZ-Code                         siehe Anhang 7, Kapitel 1
	 *                 Byte 2           Funktionsbyte
	 *
	 *                 Der Inhalt des Funktionsbytes entspricht der verallgemeinerten Definition des
	 *                 Funktionsbytes in Kapitel 6.2.18.
	 * </pre>
	 * Die folgende Tabelle stellt übersichtlich alle Varianten des Telegramms zusammen, erläutert deren Funktion und gibt
	 * an, für welche Kombination aus Anzeige- und Steuerungsprinzip die jeweilige Verwendung zulässig ist. Dabei sind
	 * Felder mit --  hinterlegt, wenn sie im betreffenden Kontext nicht relevant sind. Mit - hinterlegte Felder in der
	 * Spalte "Steuerungsprinzip" bedeuten, dass die jeweiligen Stellbefehle negativ quittiert werden, die Nummer in
	 * Klammern gibt den zu verwendenden Fehlercode an. Felder mit einem (zul) kennzeichnen die Zulässigkeit des
	 * entsprechenden Telegramms für das jeweilige DE. Eine hinter zul angefügte Nummer gekennzeichnetes Feld verweist auf
	 * einen Telegrammtyp aus früheren Versionen der TLS, die bei Altlagen im jeweiligen Kontext noch zulässig sind.<br>
	 * <pre>
	 * Steuerungsprinzip
	 * ID   Typ   Länge  Stellcode  n      Funktion                       Anzeige      f     Steuerungsprinzip
	 *                                                                    prinzip
	 *                                                                                       1        2       3
	 * 3R   33    5      0          --     Grundeinstellung pro DE        a,b          --    --       -(43)   --
	 *                              --     individuell festlegen          c-e          --    -(44)    -(43)   -(44)
	 *                              --                                    Cl           --    zul      -(43)   zul
	 *                   1-254      --     Grundeinstellung festlegen     a,b          --    --       zul 32  --
	 *                              --     mit Codenummer                 c,d          --    -(45)    zul     zul
	 *                              --                                    e,Cl         --    -(45)    zul 42  zul 42
	 *                   255        --     Grundeinstellung "son't care"  alle         --    zul      zul 42  zul 42
	 *            6+n*f  0          1 --   Grundeinstellung festlegen     a,b          --    --       -(43)   --
	 *                              1 213  mit freier Definition          c,d          1     zul      -(43)   zul
	 *                              1 106                                 e            2     zul      -(43)   zul
	 *                              1 --                                  Cl           --    --       -(43)   --
	 * 3R         5      0          --     Grundeinstellung pro DE        a,b          --    --       --      --
	 *                              --     individuell festlegen          Cl           --    zul      --      zul
	 *                   1-254      --     Grundeinstellung melden        a,b          --    --       zul 32  --
	 *                              --     mit Codenummer                 c,d          --    --       zul     zul
	 *                              --                                    e,Cl         --    --       zul 42  zul 42
	 *                   255        --     Grundeinstellung "son't care"  alle         --    zul      zul 42  zul 42
	 *            6+n*f  0          1 --   Grundeinstellung melden        a,b          --    --       --   --
	 *                              1 213  mit freier Definition          c,d          1     zul      --      zul
	 *                              1 106                                 e            2     zul      --      zul
	 *                              1 --                                  Cl           --    --       --      --
	 * 19R        2      --         --     Abruf Grundeinstellung         alle         --    zul      zul     zul
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg004Typ033Id003AtgTlsWzgGrundEinstellungAspParameterIst(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// AnzeigePrinzip aus Telegramm --------------------------------------------------------------------------------
		int _AnzeigePrinzip_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("AnzeigePrinzip").set(_AnzeigePrinzip_);

		// Stellcode und Funktion aus Telegramm ------------------------------------------------------------------------
		int _Stellcode_ = deBlock.getDeData().readUnsignedByteAsInt();
		int _Funktion_ = deBlock.getDeData().readUnsignedByteAsInt();

		// Alle dynamischen Arrays auf Länge 0, werden abhängig von der Semantik nachfolgende auf richtige Größe gesetzt.
		data.getArray("WVZAnzeige").setLength(0);
		data.getArray("StellCode").setLength(0);
		data.getArray("TextDefinition").setLength(0);
		data.getArray("GruppenDefinition").setLength(0);

		// Abhängig vom AnzeigePrinzip Stellcode und Funktion zuordnen

		switch (_AnzeigePrinzip_) {
			case 0:
			case 1: {
				data.getArray("WVZAnzeige").setLength(1);

				// Wenn Funktion = ausgeschaltet, hat WVZ-Code keine Bedeutung. Wird deshalb für
				// Protokollausgaben auf -1 = "aus" gesetzt.
				if ((_Funktion_ & 0x01) == 0) {
					_Stellcode_ = -1;
				}
				data.getArray("WVZAnzeige").getItem(0).getUnscaledValue("Code").set(_Stellcode_);

				if ((_Funktion_ & 0x02) == 0) {
					data.getArray("WVZAnzeige").getItem(0).getItem("Blinken").getUnscaledValue("Status").setText("aus");
					data.getArray("WVZAnzeige").getItem(0).getItem("Blinken").getUnscaledValue("Dauer").setText("aus");
				}
				else {
					data.getArray("WVZAnzeige").getItem(0).getItem("Blinken").getUnscaledValue("Status").setText("ein");
					int _BlinkDauer_ = _Funktion_ >>> 4;
					if ((_BlinkDauer_ < 5) || (_BlinkDauer_ > 15)) {
						debug.warning("Blinkdauer ungültig (5-15) : " + (_Funktion_ >>> 4) + Debug.NEWLINE +
						              "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE +
						              deBlock.toString());
						return resultData;
					}
					data.getArray("WVZAnzeige").getItem(0).getItem("Blinken").getUnscaledValue("Dauer").set(_BlinkDauer_);
				}

				if ((_Funktion_ & 0x04) == 0) {
					data.getArray("WVZAnzeige").getItem(0).getItem("Status").getUnscaledValue("Fehler").setText("ok");
				}
				else {
					data.getArray("WVZAnzeige").getItem(0).getItem("Status").getUnscaledValue("Fehler").setText("Fehler");
				}
				if ((_Funktion_ & 0x08) == 0) {
					data.getArray("WVZAnzeige").getItem(0).getItem("Status").getUnscaledValue("Programm").setText("abgeschlossen");
				}
				else {
					data.getArray("WVZAnzeige").getItem(0).getItem("Status").getUnscaledValue("Programm").setText("in Arbeit");
				}
			}
			break;
			case 2:
			case 3:
			case 4:
			case 8: {
				data.getArray("StellCode").setLength(1);

				// Wenn Funktion = ausgeschaltet, hat Stell-Code keine Bedeutung. Wird deshalb für
				// Protokollausgaben auf -1 = "aus" gesetzt.
				if ((_Funktion_ & 0x01) == 0) {
					_Stellcode_ = -1;
				}
				data.getArray("StellCode").getItem(0).getUnscaledValue("Code").set(_Stellcode_);
				if ((_Funktion_ & 0x02) == 0) {
					data.getArray("StellCode").getItem(0).getItem("Blinken").getUnscaledValue("Status").setText("aus");
					data.getArray("StellCode").getItem(0).getItem("Blinken").getUnscaledValue("Dauer").setText("aus");
				}
				else {
					data.getArray("StellCode").getItem(0).getItem("Blinken").getUnscaledValue("Status").setText("ein");
					int _BlinkDauer_ = _Funktion_ >>> 4;
					if ((_BlinkDauer_ < 5) || (_BlinkDauer_ > 15)) {
						debug.warning("Blinkdauer ungültig (5-15) : " + (_Funktion_ >>> 4) + Debug.NEWLINE +
						              "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE +
						              deBlock.toString());
						return resultData;
					}
					data.getArray("StellCode").getItem(0).getItem("Blinken").getUnscaledValue("Dauer").set(_BlinkDauer_);
				}
				if ((_Funktion_ & 0x04) == 0) {
					data.getArray("StellCode").getItem(0).getItem("Status").getUnscaledValue("Fehler").setText("ok");
				}
				else {
					data.getArray("StellCode").getItem(0).getItem("Status").getUnscaledValue("Fehler").setText("Fehler");
				}
				if ((_Funktion_ & 0x08) == 0) {
					data.getArray("StellCode").getItem(0).getItem("Status").getUnscaledValue("Programm").setText("abgeschlossen");
				}
				else {
					data.getArray("StellCode").getItem(0).getItem("Status").getUnscaledValue("Programm").setText("in Arbeit");
				}
			}
			break;
			default: {
				debug.warning("Anzeigeprinzip ungültig (0-4,8) : " + _AnzeigePrinzip_ + Debug.NEWLINE +
				              "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE +
				              deBlock.toString());
				return resultData;
			}
		}
		// Ermitteln, ob Langversion vorliegt...

		if (deBlock.getDeData().length() > 3) {
			int count = deBlock.getDeData().readUnsignedByteAsInt();

			// Abhängig vom AnzeigePrinzip Stellcode und Funktion zuordnen

			switch (_AnzeigePrinzip_) {
				case 2:
				case 3: {
					data.getArray("TextDefinition").setLength(count);

					char text;
					for (int i = 0; i < count; ++i) {
						text = (char) deBlock.getDeData().readUnsignedByteAsInt();
						data.getArray("TextDefinition").getItem(i).getTextValue("Text").setText("" + text);
					}
				}
				break;
				case 4: {
					data.getArray("GruppenDefinition").setLength(count / 2);

					for (int i = 0; i < count / 2; ++i) {
						_Stellcode_ = deBlock.getDeData().readUnsignedByteAsInt();
						_Funktion_ = deBlock.getDeData().readUnsignedByteAsInt();

						// Wenn Funktion = ausgeschaltet, hat Stell-Code keine Bedeutung. Wird deshalb für
						// Protokollausgaben auf -1 = "aus" gesetzt.
						if ((_Funktion_ & 0x01) == 0) {
							_Stellcode_ = -1;
						}
						data.getArray("StellCode").getItem(i).getUnscaledValue("Code").set(_Stellcode_);

						if ((_Funktion_ & 0x02) == 0) {
							data.getArray("StellCode").getItem(i).getItem("Blinken").getUnscaledValue("Status").setText("aus");
							data.getArray("StellCode").getItem(i).getItem("Blinken").getUnscaledValue("Dauer").setText("aus");
						}
						else {
							data.getArray("StellCode").getItem(i).getItem("Blinken").getUnscaledValue("Status").setText("ein");
							int _BlinkDauer_ = _Funktion_ >>> 4;
							if ((_BlinkDauer_ < 5) || (_BlinkDauer_ > 15)) {
								debug.warning("Blinkdauer ungültig (5-15) : " + (_Funktion_ >>> 4) + Debug.NEWLINE +
								              "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE +
								              deBlock.toString());
								return resultData;
							}
							data.getArray("StellCode").getItem(i).getItem("Blinken").getUnscaledValue("Dauer").set(_BlinkDauer_);
						}

						if ((_Funktion_ & 0x04) == 0) {
							data.getArray("StellCode").getItem(i).getItem("Status").getUnscaledValue("Fehler").setText("ok");
						}
						else {
							data.getArray("StellCode").getItem(i).getItem("Status").getUnscaledValue("Fehler").setText("Fehler");
						}
						if ((_Funktion_ & 0x08) == 0) {
							data.getArray("StellCode").getItem(i).getItem("Status").getUnscaledValue("Programm").setText("abgeschlossen");
						}
						else {
							data.getArray("StellCode").getItem(i).getItem("Status").getUnscaledValue("Programm").setText("in Arbeit");
						}
					}
				}
				break;
				default: {
					debug.warning("Anzeigeprinzip ungültig (0-4,8) : " + _AnzeigePrinzip_ + Debug.NEWLINE +
					              "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE +
					              deBlock.toString());
					return resultData;
				}
			}
		}
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}
}
